/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#pragma once

#include <ox/std/string.hpp>
#include <ox/std/types.hpp>
#include <ox/std/memory.hpp>

#include "fieldcounter.hpp"
#include "optype.hpp"

namespace ox {

struct TypeNameCatcher {

	const char *name = "";

	constexpr TypeNameCatcher() noexcept = default;

	template<typename T = std::nullptr_t>
	constexpr void setTypeInfo(const char *n = T::TypeName, int = 0) noexcept {
		this->name = n;
	}

	template<typename T>
	constexpr Error field(const char*, T*, std::size_t) noexcept {
		return OxError(0);
	}

	template<typename T>
	constexpr Error field(const char*, T) noexcept {
		return OxError(0);
	}

	static constexpr auto opType() noexcept {
		return OpType::WriteDefinition;
	}

};

template<typename T, typename Str = const char*>
constexpr Str getModelTypeName() noexcept {
	auto a = std::allocator<T>();
	auto t = a.allocate(1);
	TypeNameCatcher nc;
	oxIgnoreError(model(&nc, t));
	a.deallocate(t, 1);
	return nc.name;
}

template<typename T, typename Str = const char*>
constexpr Str getModelTypeName(T *val) noexcept {
	TypeNameCatcher nc;
	oxIgnoreError(model(&nc, val));
	return nc.name;
}

template<typename T>
consteval auto requireModelTypeName() noexcept {
	constexpr auto name = getModelTypeName<T>();
	static_assert(ox_strcmp(name, "") != 0, "TypeName is required");
	return name;
}

template<typename T, typename Str = const char*>
constexpr auto ModelTypeName_v = getModelTypeName<T, Str>();

}
